/*
 * @(#)file      Agent.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.7
 * @(#)lastedit  04/04/21
 * 
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

import java.net.MalformedURLException;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Set;
import java.util.Iterator;
import java.io.IOException;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.MBeanServerFactory;
import javax.management.InstanceNotFoundException;

import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXServiceURL;

import com.sun.management.comm.SnmpV3AdaptorServer;
import com.sun.jdmk.comm.CommunicatorServer;
import com.sun.management.snmp.InetAddressAcl;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpOidTable;

import com.sun.management.snmp.SnmpOidDatabaseSupport;

/**
 * This example shows how to work with SNMP table in Java DMK.
 * This class defines the main entry point for this example.
 * It creates an MBeanServer - in which it registers some simple MBeans.
 * It creates and starts connector servers for that MBeanServer, one for
 * each JmxServiceURL specified as argument to main().
 * It creates a trilingual SnmpV3AdaptorServer, in which it register an
 * instance of the JMX-MBEAN-SERVER-MIB. 
 * It creates a Timer in which it registers periodical tasks that will
 * add and remove MBeans from the MBeanServer, in order to demonstrate the
 * dynamic nature of the MIB.
 * Then it starts everything and wait forever.
 * <p><b>This API is a Sun Microsystems internal API  and is subject 
 * to change without notice.</b></p>
 **/
public class Agent {

    /**
     * The Simple MBean - just for the sake of the example 
     **/
    public interface SimpleMBean {
	public String getName();
	public int    getCount();
	public void   reset();
    }

    /**
     * An implementation of the Simple MBean - just for the sake of the 
     * example 
     **/
    public static class Simple implements SimpleMBean {
	public Simple() {
	    this(null);
	}

	public Simple(String name) {
	    final String id = makeName();
	    this.name=((name==null)?id:name); 
	    this.count=0;
	}

	public final synchronized String getName() {
	    count++;
	    return name;
	}
	public synchronized int getCount() {
	    return count;
	}

	public synchronized void reset() {
	    count=0;
	}

	public static synchronized String makeName() {
	    return "simple-"+instanceCount++;
	}

	private final String name;
	private int count;
	private static int instanceCount=0;
    }

    /**
     * Share the MBeanServer with any other application.
     *
     **/
    public static MBeanServer getPlatformMBeanServer() {
	final MBeanServer test = MBeanServerFactory.createMBeanServer();
	final MBeanServer first = (MBeanServer)
	    MBeanServerFactory.findMBeanServer(null).get(0);
	if (test != first) MBeanServerFactory.releaseMBeanServer(test);
	return first;
    }

    /**
     * Register some Simple MBeans in the MBeanServer.
     **/
    public int populate(MBeanServer server, int mbeanCount) {
	int count = 0;
	for (int i=0; i<mbeanCount; i++) {
	    try {
		final SimpleMBean simple = new Simple();
		final ObjectName name = 
		    new ObjectName("Example:type=Simple,name="+
				   simple.getName());
		server.registerMBean(simple,name);
		count ++;
		System.out.println("Registered MBean: " + name);
	    } catch (Exception x) {
		System.err.println("Failed to register MBean: " + x);
		debug(x);
	    }
	}
	return count;
    }

    /**
     * Deregister some Simple MBeans from the MBeanServer.
     **/
    public int depopulate(MBeanServer server, int mbeanCount) {
	int count = 0;
	while (true) {
	    try { 
		final ObjectName pattern = 
		    new ObjectName("Example:type=Simple,*");
		Set mbeans = server.queryNames(pattern,null);
		if (mbeans.isEmpty()) break;
		for (Iterator it=mbeans.iterator(); it.hasNext() ; ) {
		    if (count == mbeanCount) return count;
		    final ObjectName next = (ObjectName) it.next();
		    try {
			server.unregisterMBean(next);
			count++;
			System.out.println("Deregistered MBean: " + next);
		    } catch (InstanceNotFoundException x) {
			continue;
		    }
		}
		if (count >= mbeanCount) break;
	    } catch (Exception x) {
		System.err.println("Unexpected exception: " + x);
		debug(x);
		break;
	    }
	}
	return count;
    }

    /**
     * Create and start the SnmpV3AdaptorServer and the JMX-MBEAN-SERVER-MIB.
     **/
    public void startmib(MBeanServer server,int port) throws IOException {
	final SnmpV3AdaptorServer adaptor = 
	    new SnmpV3AdaptorServer((InetAddressAcl)null,port);

	adaptor.enableSnmpV1V2SetRequest();
	adaptor.start();
	do {
	    adaptor.waitState(CommunicatorServer.ONLINE,1000);
	} while (adaptor.getState() == CommunicatorServer.STARTING);

	final int state = adaptor.getState();
	if (state != CommunicatorServer.ONLINE) {
	    try { adaptor.stop(); } catch (Exception x) { /* OK */ }
	    throw new IOException("Can't start adaptor: " + 
				  adaptor.getStateString());
	}

	JMX_MBEAN_SERVER_MIB_Impl mib = new JMX_MBEAN_SERVER_MIB_Impl(server);

	try {
	    mib.init();
	    mib.setSnmpAdaptor(adaptor);
	    server.registerMBean(adaptor,null);
	    mib.start();
	} catch (Exception x) {
	    System.err.println("Failed to register SnmpAdaptor: " + x);
	    try { adaptor.stop(); } catch (Exception e) { /* OK */ }
	    final IOException io = 
		new IOException("Failed to register SnmpAdaptor");
	    io.initCause(x);
	    throw io;
	}

	System.out.println("SnmpAdaptor ready at port: " + adaptor.getPort());
    }

    /**
     * Creates and starts everything. Parameters are expected to be 
     * JMXServiceURL strings, e.g. "service:jmx:jmxmp://".
     * One JMXConnectorServer will be created from each given string in args.
     * Creates and populate the MBeanServer, creates and starts 
     * JMXConnectorServer if any URL is specified, creates and start the
     * SnmpV3AdaptorServer and the JMX-MBEAN-SERVER-MIB, then creates
     * the timer tasks.
     **/
    public void start(int port, String[] args) throws IOException {
	final MBeanServer server =  getPlatformMBeanServer();
	final JMXServiceURL[] urls = new JMXServiceURL[args.length];
	for (int i=0;i<args.length;i++) {
	    try {
		urls[i] = new JMXServiceURL(args[i]);
	    } catch (MalformedURLException x) {
		System.err.println("Invalid parameter: " + args[i]);
		System.err.println("Syntax is: java " + 
				   this.getClass().getName() +
				   " [JMXServiceURL-1, JMXServiceURL-2, ...]");
		throw x;
	    }
	}

	for (int i=0;i<urls.length;i++) {
	    try {
		final JMXConnectorServer s = 
		    JMXConnectorServerFactory.newJMXConnectorServer(urls[i],
								    null,null);
		final ObjectName name = 
		    new ObjectName("Connector:type="+s.getClass().getName()+
				   ",instance="+i);
		server.registerMBean(s,name);
		s.start();
		System.out.println(name + ": started.");
		System.out.println("Connector ready at: " + s.getAddress());
	    } catch (Exception x) {
		final String msg = "Failed to start connector: " + args[i];
		System.err.println(msg);
		final IOException io = new IOException(msg);
		io.initCause(x);
		throw io;
	    }
	}

	populate(server,5);

	startmib(server,port);
	
	final Timer timer = new Timer();
	final long  period= 2000;
	final long  now=0;
	final int   max=4;
	for (int i=0;i<max;i++) {
	    final TimerTask taskAdd = new TimerTask() {
		    public void run() {
			final MBeanServer server =  getPlatformMBeanServer();
			populate(server,1);
		    }
		};
	    final TimerTask taskRem = new TimerTask() {
		    public void run() {
			final MBeanServer server =  getPlatformMBeanServer();
			depopulate(server,1);
		    }
		};
	    timer.schedule(taskAdd,now+period*i,2*period*max);
	    timer.schedule(taskRem,now+period*max+period*i,2*period*max);
	}
    }

    /**
     * Initialize the SNMP OID table.
     **/
    public static void initOidTable() {	
	final SnmpOidTable orig = SnmpOid.getSnmpOidTable();
	SnmpOidDatabaseSupport mibs = new SnmpOidDatabaseSupport();
	mibs.add(new JMX_MBEAN_SERVER_MIBOidTable());
	if (orig != null) mibs.add(orig);
	SnmpOid.setSnmpOidTable(mibs);
    }

    /**
     * Main entry point. Parameters are expected to be 
     * JMXServiceURL strings, e.g. "service:jmx:jmxmp://".
     * One JMXConnectorServer will be created from each given string in args.
     **/
    public static void main(String[] args) {

	Agent agent = new Agent();
	try {
	    initOidTable();
	    final String portStr = System.getProperty("snmp.port","16161");
	    final int port;
	    try {
		port = Integer.parseInt(portStr);
	    } catch (Exception x) {
		System.err.println("Invalid value specified for snmp.port: "
				   + portStr);
		System.err.println("Error is: " + x);
		throw x;
	    }
	    System.out.println("Using SNMP port " + port);

	    agent.start(port,args);

	    // Wait forever...
	    Thread.sleep(Long.MAX_VALUE);
	} catch (Exception x) {
	    System.err.println("Failed to start agent: " + x);
	    debug(x);
	    System.exit(1);
	}

    }

    public static void debug(Throwable x) {
	if (Boolean.getBoolean("debug")) x.printStackTrace();
    }
}
