/*
 * @(#)file      JmxMBeanTableMetaImpl.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.6
 * @(#)lastedit  04/04/21
 * 
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// java imports
//
import java.io.Serializable;

import java.util.Vector;
import java.util.Map;
import java.util.TreeMap;
import java.util.Collection;
import java.util.Comparator;
import java.util.Arrays;
import java.util.Iterator;

import java.lang.ref.WeakReference;


// jmx imports
//
import javax.management.MBeanServer;
import javax.management.ObjectName;
import com.sun.management.snmp.SnmpCounter;
import com.sun.management.snmp.SnmpCounter64;
import com.sun.management.snmp.SnmpGauge;
import com.sun.management.snmp.SnmpInt;
import com.sun.management.snmp.SnmpUnsignedInt;
import com.sun.management.snmp.SnmpIpAddress;
import com.sun.management.snmp.SnmpTimeticks;
import com.sun.management.snmp.SnmpOpaque;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.SnmpStringFixed;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpNull;
import com.sun.management.snmp.SnmpValue;
import com.sun.management.snmp.SnmpVarBind;
import com.sun.management.snmp.SnmpStatusException;

// jdmk imports
//
import com.sun.management.snmp.agent.SnmpIndex;
import com.sun.management.snmp.agent.SnmpMib;
import com.sun.management.snmp.agent.SnmpMibTable;
import com.sun.management.snmp.agent.SnmpMibSubRequest;
import com.sun.management.snmp.agent.SnmpStandardObjectServer;
import com.sun.management.snmp.EnumRowStatus;

/**
 * This class is used for implementing the "JmxMBeanTable" meta.
 *
 * It extends the generated JmxMBeanTableMeta class in order to override 
 * the following methods:
 *
 *   - getEntry(SnmpOid rowOid): returns an JmxMBeanEntryImpl object
 *     for the indicated row.
 *   - contains(SnmpOid oid, Object userData): Tells whether the indicated
 *     row is present int the table.
 *   - getNextOid(Object userData): returns the rowOid (row index) of the
 *     first row in the table.
 *   - getNextOid(SnmpOid rowOid,Object userData): returns the index of 
 *     the row that immediately follows the indicated row in the table.
 *   - createNewEntry(SnmpMibSubRequest request, SnmpOid rowOid, int depth):
 *     makes it possible to create new rows from SNMP SET (overriding
 *     this method is not necessary when the table is read-only)
 *   - removeTableRow(SnmpMibSubRequest request, SnmpOid rowOid, int depth):
 *     makes it possible to delete rows from SNMP SET (overriding
 *     this method is not necessary when the table is read-only)
 * 
 * This class also shows how to implement a two-levels caching mechanism,
 * taking benefit of the SnmpUserData factory in order to populate a 
 * transient request-contextual cache (the request contextual cache is
 * garbage collected at the end of the request):
 *
 *   - The JmxMBeanTableData is a snapshot of the table index. It is 
 *     computed on demand, and cached in a JmxMBeanTableCache object
 *     which holds it through a WeakReference. The JmxMBeanTableData 
 *     includes a time-stamp which is used to determined whether the
 *     cached data is still valid, or if it needs to be recomputed.
 *
 *   - The JmxMBeanTableCache is a weak-cache used to avoid constant
 *     recomputation of the JmxMBeanTableData. It holds the last computed
 *     JmxMBeanTableData in a weak reference, so that the garbage collector
 *     might reclaim it at any time (provided that no request is currently
 *     working on that last computed copy).
 *     When an SNMP request that needs to access the jmxMBeanTable comes
 *     in, the JmxMBeanTableData is looked-up from the JmxMBeanTableCache,
 *     and recomputed if needed. The JmxMBeanTableData object is then
 *     put in the request-contextual cache where it remains until the
 *     request is completed. The request will always consult that copy,
 *     ensuring the consistency of the returned data. If in the mean
 *     time a concurrent request comes in, it might obtain the same
 *     JmxMBeanTableData copy than the previous request, or a newly
 *     computed object, if the previous one was found obsolete. So the
 *     two requests might work on the same, or different snapshot of
 *     the table. The snapshot obtained by a given request never changes.
 *     A new request might obtain a new snapshot.
 *
 *   - The JmxMBeanTableDataItem is a simple class that wraps an 
 *     ObjectName and a JmxMBeanEntry. The JmxMBeanEntry is lazy-evaluated.
 *     The JmxMBeanTableData contains an array of JmxMBeanTableDataItem.
 **/
public class JmxMBeanTableMetaImpl extends JmxMBeanTableMeta {

    /**
     * Compares two SnmpOid.
     **/
    public static final Comparator oidComparator = new Comparator() {
	    public int compare(Object o1, Object o2) {
		return ((SnmpOid)o1).compareTo((SnmpOid)o2);
	    }
	    public boolean equals(Object o1, Object o2) {
		if (o1 == o2) return true;
		else return o1.equals(o2);
	    }
	};


    /**
     * Compares two ObjectNames.
     **/
    public static final Comparator objectNameComparator = new Comparator() {
	    public int compare(Object o1, Object o2) {
		return ((ObjectName)o1).getCanonicalName().
		    compareTo(((ObjectName)o2).getCanonicalName());
	    }
	    public boolean equals(Object o1, Object o2) {
		if (o1 == o2) return true;
		else return o1.equals(o2);
	    }
	};


    /**
     * Temporary object kept in the table cache.
     * Makes it possible to do a lazy evaluation of the MBean entry.
     **/
    public static class JmxMBeanTableDataItem {
	private final ObjectName name;
	public JmxMBeanTableDataItem(ObjectName name) {
	    this.name  = name;
	    this.entry = null;
	}
	public ObjectName getObjectName() {
	    return name;
	}
	public String toString() {
	    return name.toString();
	}
	public synchronized JmxMBeanEntryImpl getEntry() {
	    return entry;
	}
	private synchronized void setEntry(JmxMBeanEntryImpl entry) {
	    this.entry = entry;
	}
	private JmxMBeanEntryImpl entry;
    }

    /**
     * JmxMBeanTable data contained in the JmxMBeanTable cache. 
     * When a request is made on the JmxMBeanTable, we look whether there
     * are some valid data in the table cache. If there are, we take them.
     * If not we compute a new JmxMBeanTable data object.
     * The JmxMBeanTable data is a sorted array of JmxMBeanTableDataItem, 
     * associated with their rowOid (index part).
     * An instance of JmxMBeanTableData can be shared by several concurrent
     * request. It does not need to be synchronized because it is immutable.
     * It is a snapshot of the table index, taken at the `lastUpdated' time
     * stamp.
     * The JmxMBeanTableData is stored in the request contextual cache for
     * the duration of the request in order to ensure a consistent behavior
     * for the duration of the request. 
     * When all the request holding it have completed it may be reclaimed
     * by the garbage collector. See JmxMBeanTableCache below.
     **/
    public static class JmxMBeanTableData {
	/**
	 * Constructs a new instance of JmxMBeanTableData. Instances are
	 * immutable.
	 * @param lastUpdated Time stamp as returned by 
	 *        {@link System#currentTimeMillis System.currentTimeMillis()}
	 * @param indexes The table entry indexes, sorted in ascending order.
	 * @param datas   The table datas, sorted according to the 
	 *                order in <code>indexes<code>: <code>datas[i]</code>
	 *                is the data that corresponds to 
	 *                <code>indexes[i]</code>
	 **/
	public JmxMBeanTableData(long lastUpdated, SnmpOid indexes[], 
			      Object  datas[]) {
	    this.lastUpdated = lastUpdated;
	    this.indexes     = indexes;
	    this.datas       = datas;
	}

	/**
	 * Constructs a new instance of JmxMBeanTableData. Instances are
	 * immutable.
	 * @param lastUpdated Time stamp as returned by 
	 *        {@link System#currentTimeMillis System.currentTimeMillis()}
	 * @param indexMap The table indexed table data, sorted in ascending 
	 *                 order by {@link #oidComparator}. The keys must be
	 *                 instances of {@link SnmpOid}.
	 **/
	public JmxMBeanTableData(long lastUpdated, TreeMap indexMap) {
	    this(lastUpdated, indexMap, true);
	}

	/**
	 * Constructs a new instance of JmxMBeanTableData. Instances are
	 * immutable.
	 * @param lastUpdated Time stamp as returned by 
	 *        {@link System#currentTimeMillis System.currentTimeMillis()}
	 * @param indexMap The table indexed table data, sorted in ascending 
	 *                 order by {@link #oidComparator}. The keys must be
	 *                 instances of {@link SnmpOid}.
	 **/
	public JmxMBeanTableData(long lastUpdated, TreeMap indexMap, 
				 boolean b) {

	    final int size = indexMap.size();
	    this.lastUpdated = lastUpdated;
	    this.indexes     = new SnmpOid[size];
	    this.datas       = new Object[size];
	    
	    if(b) {
		indexMap.keySet().toArray(this.indexes);	
		indexMap.values().toArray(this.datas);
	    } else
		indexMap.values().toArray(this.datas);
	}

	/**
	 * Time stamp as returned by 
	 * {@link System#currentTimeMillis System.currentTimeMillis()}
	 **/
	public final long    lastUpdated;
    
	/**
	 * The table entry indexes, sorted in ascending order.
	 **/
	public final SnmpOid indexes[];
	
	/**
	 * The table datas, sorted according to the 
	 * order in <code>indexes<code>: <code>datas[i]</code>
	 * is the data that corresponds to <code>indexes[i]</code>
	 **/
	public final Object  datas[];
    
	/**
	 * The position of the given <var>index</var>, as returned by
	 * <code>java.util.Arrays.binarySearch()</code>
	 **/
	public final int find(SnmpOid index) {
	    return Arrays.binarySearch(indexes,index,oidComparator);
	}
    
	// JmxMBeanTableData.getData()
	public  Object getData(SnmpOid index) {
	    final int pos = find(index);
	    if ((pos < 0)||(pos >= datas.length)) return null;
	    return datas[pos];
	}

	// JmxMBeanTableData.getNext()
	public  SnmpOid getNext(SnmpOid index) {
	    if (index == null) {
		if (indexes.length>0) return indexes[0];
		else return null;
	    }
	    final int pos = find(index);
	    if (pos > -1) {
		if (pos < (indexes.length -1) ) return indexes[pos+1]; 
		else return null;
	    }
	    final int insertion = -pos -1;
	    if ((insertion > -1) && (insertion < indexes.length))
		return indexes[insertion];
	    else return null;
	}

	// JmxMBeanTableData.contains()
	public  boolean contains(SnmpOid index) {
	    final int pos = find(index);
	    return ((pos > -1)&&(pos < indexes.length));
	}

	public int size() {
	    return indexes.length;
	}
    }

    /**
     * The JmxMBeanTable Cache. This object holds a WeakReference on a
     * JmxMBeanTableData object. It is responsible for computing a new
     * JmxMBeanTableData object when needed.
     * When a new request needs to access the JmxMBeanTable, the 
     * JmxMBeanTableCache is asked for a JmxMBeanTableData object.
     * If the WeakReference holding the last computed JmxMBeanTableData object 
     * has not been reclaimed by the garbage collector, and if that object
     * is still valid, the JmxMBeanTableCache returns it. Otherwise,
     * it computes a new JmxMBeanTableData object.
     * 
     **/
    public static class JmxMBeanTableCache {

	/**
	 * Interval of time in ms during which the cached table data
	 * is considered valid.
	 **/
	private long validity;

	/**
	 * A weak reference holding cached table data.
	 **/
	private transient WeakReference datas;

	/**
	 * This map associate an entry name with the SnmpOid index that's
	 * been allocated for it. It is updated every time the cached datas
	 * are updated.
	 **/
	private TreeMap names = new TreeMap(objectNameComparator);
    
	/**
	 * The last allocate index.
	 **/
	private long last = 0;

	/**
	 * true if the index has wrapped.
	 **/
	private boolean   wrapped = false;

	final JmxMBeanTableMetaImpl     meta;
	final JMX_MBEAN_SERVER_MIB_Impl myMib;

	JmxMBeanTableCache(JmxMBeanTableMetaImpl meta,
			   JMX_MBEAN_SERVER_MIB_Impl myMib) {
	    this.validity = myMib.getCacheValidity();
	    this.meta     = meta;
	    this.myMib    = myMib;
	}

	public MBeanServer getMibServer() {
	    return  myMib.getMibServer();
	}

	/**
	 * Returns the cached table data, if it is still valid,
	 * or recompute it if it is obsolete. 
	 * <p>
	 * When cache data is recomputed, store it in the weak reference,
	 * unless {@link #validity} is 0: then the data will not be stored
	 * at all.<br>
	 * This method calls {@link #isObsolete(JmxMBeanTableData)} to 
	 * determine whether the cached data is obsolete, and {
	 * {@link #updateCachedDatas(Object)} to recompute it.
	 * </p>
	 * @return the valid cached data, or the recomputed table data.
	 **/
	public synchronized JmxMBeanTableData getTableDatas() {
	    final Map userData = JmxMBeanContextFactory.getUserData();
	    return getTableDatas(userData);
	}

	/**
	 * Returns the cached table data, if it is still valid,
	 * or recompute it if it is obsolete. 
	 * <p>
	 * When cache data is recomputed, store it in the weak reference,
	 * unless {@link #validity} is 0: then the data will not be stored
	 * at all.<br>
	 * This method calls {@link #isObsolete(JmxMBeanTableData)} to 
	 * determine  whether the cached data is obsolete, and 
	 * {@link #updateCachedDatas(Object)} to recompute it.
	 * </p>
	 * @param userData The request contextual cache allocated by the
	 *        {@link JmxMBeanContextFactory}.
	 * @return the valid cached data, or the recomputed table data.
	 **/
	public synchronized JmxMBeanTableData getTableDatas(Object userData) {
	    final JmxMBeanTableData cached   = getCachedDatas();
	    if (cached != null) return cached;
	    final JmxMBeanTableData computedDatas = 
		updateCachedDatas(userData);
	    if (validity != 0) datas = new WeakReference(computedDatas);
	    return computedDatas;
	}

	/**
	 * Returns the cached table data.
	 * Returns null if the cached data is obsolete, or if there is no
	 * cached data, or if the cached data was garbage collected.
	 * @return a still valid cached data or null. 
	 **/
	private JmxMBeanTableData getCachedDatas() {
	    if (datas == null) return null;
	    final JmxMBeanTableData cached = (JmxMBeanTableData) datas.get();
	    if ((cached == null) || isObsolete(cached)) return null;
	    return cached;
	}

	/**
	 * true if the given cached table data is obsolete.
	 **/
	private boolean isObsolete(JmxMBeanTableData cached) {
	    if (cached   == null) return true;
	    if (validity < 0)     return false;
	    return ((System.currentTimeMillis() - cached.lastUpdated) 
		    > validity);
	}

	/**
	 * Update cached datas.
	 * Obtains a {@link Collection} of raw datas by calling 
	 * {@link #getMBeanNames(Map) getMBeanNames((Map)userData)}.<br>
	 * Then allocate a new {@link TreeMap} to serve as temporary map 
	 * between names and indexes, and update the cached table data.<br>
	 * Finally replaces the {@link #names} TreeMap by the temporary
	 * TreeMap.
	 * @param userData The request contextual cache allocated by the
	 *        {@link JmxMBeanContextFactory}.
	 **/
	private JmxMBeanTableData updateCachedDatas(Object userData) {

	    final Map userDataMap = 
		(userData instanceof Map)?(Map)userData:null;

	    // Look for memory manager list in request contextual cache.
	    final Collection mbeanNames = getMBeanNames(userDataMap);

	    if (Boolean.getBoolean("debug"))
		System.err.println("updateCachedDatas: mbeanNames.size()=" + 
		       ((mbeanNames==null)?"<no data>":""+mbeanNames.size()));

	    TreeMap ctxt = new TreeMap(objectNameComparator);

	    final int size = ((mbeanNames == null)?0:mbeanNames.size());
	    if (size == 0) return  null;

	    final long time = System.currentTimeMillis();
	    final Iterator it  = mbeanNames.iterator();
	    final TreeMap  map = new TreeMap(oidComparator);
	    for (int rank=0; it.hasNext() ; rank++) {
		final ObjectName  item  = (ObjectName) it.next();
		final SnmpOid index = getIndex(ctxt,item);
		if (index == null) continue;
		map.put(index,makeTableDataItem(item));
	    }

	    final JmxMBeanTableData result = new JmxMBeanTableData(time,map);
	    names = ctxt;
	    return result;
	}

	/**
	 * Make a new instance of a JmxMBeanTablaDataItem to store in the
	 * cache. At first, the JmxMBeanTablaDataItem only wraps an 
	 * ObjectName. If the request does not need to actually access to
	 * the data exposed for that MBean, then JmxMBeanTableDataItem 
	 * remains unchanged. Otherwise, a JmxMBeanEntryImpl object can
	 * be computed on the fly and temporarily cached in the 
	 * JmxMBeanTableDataItem. The JmxMBeanEntryImpl object will be
	 * reclaimed by the garbage collector when the JmxMBeanTableDataItem
	 * is reclaimed.
	 **/
	private Object makeTableDataItem(ObjectName item) {
	    return new JmxMBeanTableDataItem(item);
	}

	/**
	 * Get a list of raw data from which to build the cached data.
	 * Obtains a list of raw data by first looking it up in the
	 * request contextual cache <var>userData</var>. 
	 * If nothing is found in the cache, calls
	 * <tt>getMibServer().queryNames(null,null)</tt> in order to obtain 
	 * a new  mbeanName list, and cache the result in <var>userData</var>.
	 * @param userData The request contextual cache allocated by
	 *        the {@link JmxMBeanContextFactory}.
	 *
	 **/
	private final Collection getMBeanNames(Map userData) {
	    Collection mbeanNames = null;

	    final String key = "JmxMBeanTable.queryNames";

	    // Look for memory manager list in request contextual cache.
	    if (userData != null) 
		mbeanNames = (Collection) userData.get(key);
	    
	    if (mbeanNames == null) {
		// No list in contextual cache, get it from API
		try {
		    mbeanNames = getMibServer().queryNames(null,null);
		} catch (Exception x) {
		    if (Boolean.getBoolean("info")) {
			System.err.println("Failed to get list of MBeans: "+
					   x);
		    }
		    if (Boolean.getBoolean("debug")) {
			x.printStackTrace();
		    }
		}

		// Put list in cache...
		if (mbeanNames != null && userData != null) 
		    userData.put(key, mbeanNames);
	    }

	    return mbeanNames;
	}

	/**
	 * Find a new index for the entry corresponding to the 
	 * given <var>item</var>.
	 * <br>This method is called by {@link #getIndex(Map,ObjectName)}
	 * when a new index needs to be allocated for an <var>item</var>. The 
	 * index returned must not be already in used.
	 * @param nameMap The nameMap allocated by 
	 *        {@link #updateCachedDatas(Object)}.
	 * @param item The ObjectName for which an index must be 
	 *        determined.
	 **/
	private SnmpOid makeIndex(Map nameMap, ObjectName item) {
	
	    // check we are in the limits of positive 32.
	    if (++last > 0x000000007FFFFFFFL) {
		// we just wrapped. 
		if (Boolean.getBoolean("debug"))
		    System.err.println("makeIndex: Index wrapping...");
		last = 0; 
		wrapped=true;
	    }

	    // If we never wrapped, we can safely return last as new index.
	    if (!wrapped) return new SnmpOid(last);

	    // We wrapped. We must look for an unused index. 
	    for (int i=1;i < 0x000000007FFFFFFFL;i++) {
		if (++last >  0x000000007FFFFFFFL) last = 1;
		final SnmpOid testOid = new SnmpOid(last);

		// Was this index already in use?
		if (names == null) return testOid;
		if (names.containsValue(testOid)) continue;

		// Have we just used it in a previous iteration?
		if (nameMap == null) return testOid;
		if (nameMap.containsValue(testOid)) continue;

		// Ok, not in use.
		return testOid;
	    }
	    // all indexes are in use! we're stuck.
	    // // throw new IndexOutOfBoundsException("No index available.");
	    // better to return null and log an error.
	    return null;
	}

	/**
	 * Check whether the index provided by the remote SNMP manager is
	 * suitable for creating a new MBean from remote. If so return
	 * the long value for that index. Otherwise, throws an exception.
	 **/
	public synchronized long checkIndex(SnmpOid oid) 
	    throws SnmpStatusException {
	    if (oid == null || oid.getLength()<1) {
		if (Boolean.getBoolean("info"))
		    System.err.println("Index is not suitable: " + oid);
		throw new SnmpStatusException(
		      SnmpStatusException.snmpRspInconsistentValue);
	    }
	    if (names != null && names.containsValue(oid)) {
		if (Boolean.getBoolean("info"))
		    System.err.println("Index is not suitable: " + oid);
		throw new SnmpStatusException(
                           SnmpStatusException.snmpRspInconsistentValue);
	    }
	    final long x = oid.getOidArc(0);
	    if (x > 0x000000007FFFFFFFL) {
		if (Boolean.getBoolean("info"))
		    System.err.println("Index is not suitable: " + oid);
		throw new SnmpStatusException(
                           SnmpStatusException.snmpRspInconsistentValue);
	    }
	    if (x > last) last = x;
	    return x;
	}

	/**
	 * Called when a remote SNMP manager tries to create a new MBean
	 * through a SET operation. This is necessary to ensure a good
	 * index management policy. This method creates a new association
	 * between an ObjectName and its associated SnmpOid index.
	 * The caller must take care to lock this cache object until the
	 * MBean by that ObjectName is successfully registered in the 
	 * MBeanServer.
	 **/
	public synchronized void registerName(ObjectName name, SnmpOid index) 
	    throws SnmpStatusException {

	    if ((index == null) || (name == null)) {
		if (Boolean.getBoolean("info"))
		    System.err.println("Index is not suitable: " + index);
		throw new SnmpStatusException(
                           SnmpStatusException.snmpRspInconsistentValue);
	    }

	    names.put(name,index);

	    if (Boolean.getBoolean("debug"))
		System.err.println("registerName: key="+name+
				   ", index="+index);
	}

	/**
	 * Check whether an index was already allocated
	 * for the entry by that name. If yes return it. Otherwise, call
	 * {@link #makeIndex(Map,ObjectName)} to compute a new
	 * index for that entry.
	 * Finally store the association between
	 * the name and index in the context TreeMap.
	 * @param nameMap The index Map allocated by
	 *        {@link #updateCachedDatas(Object)}. 
	 * @param item The ObjectName for which an index must be 
	 *        determined.
	 **/
	private SnmpOid getIndex(Map nameMap, ObjectName item) {
	    final Object index = 
		(names==null||item==null)?null:names.get(item);
	    final SnmpOid result = 
		((index != null)?((SnmpOid)index):makeIndex(nameMap,item));
	    if ((nameMap != null) && (item != null) && (result != null))
		((Map)nameMap).put(item,result);
	    if (Boolean.getBoolean("debug"))
		System.err.println("getIndex: key="+item+", index="+result);
	    return result;
	}


    }

    // The weak cache for this table.
    //
    final JmxMBeanTableCache cache; 

    // The JMX MBEAN MIB
    //
    final JMX_MBEAN_SERVER_MIB_Impl myMib;

    /**
     * Constructor for the table. Initialize metadata for "JmxMBeanTableMeta".
     */
    public JmxMBeanTableMetaImpl(SnmpMib myMib, 
				 SnmpStandardObjectServer objserv) {
        super(myMib,objserv);
	this.myMib = (JMX_MBEAN_SERVER_MIB_Impl)myMib;
	this.cache = 
	    new JmxMBeanTableCache(this,(JMX_MBEAN_SERVER_MIB_Impl)myMib);
	setCreationEnabled(true);
    }



    // Returns the rowOid (row index) of the first row in the table.
    // See com.sun.jmx.snmp.agent.SnmpMibTable
    //
    protected SnmpOid getNextOid(Object userData) 
	throws SnmpStatusException {
	// null means get the first OID.
	return getNextOid(null,userData);
    }

    // Returns the index of the row that immediately follows the
    // indicated row in the table.
    // See com.sun.jmx.snmp.agent.SnmpMibTable
    //
    protected SnmpOid getNextOid(SnmpOid oid, Object userData)
	throws SnmpStatusException {
	final boolean dbg = Boolean.getBoolean("debug");
	if (dbg) System.err.println("getNextOid: previous=" + oid);
	

	// Get the cached data.
	//
	JmxMBeanTableData tableData = getTableDatas(userData);
	if (tableData == null) {
	    // This should never happen.
	    // If we get here it's a bug.
	    //
	    if (dbg) System.err.println("getNextOid: tableData is null!");
	    throw new SnmpStatusException(SnmpStatusException.noSuchInstance);
	}
	
	// Get the next oid
	//
	SnmpOid next = oid;

	// Loop until we manage to pre-load the MBean data of the next.
	//
	do {
	    next = tableData.getNext(next);
	    if (dbg) System.err.println("getNextOid: next=" + next);

	    // if next is null: we reached the end of the table.
	    //
	    if (next == null) 
		throw new 
		    SnmpStatusException(SnmpStatusException.noSuchInstance);

	} while (loadInCache(userData,tableData,next) == null);
	
	return next;
    }
    

    // Tells whether the indicated row is present int the table.
    // See com.sun.jmx.snmp.agent.SnmpMibTable
    //
    protected boolean contains(SnmpOid oid, Object userData) {

	// Get the tableData.
	//
	JmxMBeanTableData tableData = getTableDatas(userData);

	// tableData should never be null.
	//
	if (tableData == null) 
	    return false;

	if (tableData.contains(oid)) try {
	    // Preload the MBean Data in cache.
	    //
	    return (loadInCache(userData,tableData,oid) != null);
	} catch (Exception x) {
	    if (Boolean.getBoolean("info")) {
		System.err.println("Failed to load MBean entry for " +
				   oid + ": "+ x);
	    }
	    if (Boolean.getBoolean("debug")) {
		x.printStackTrace();
	    }
	}

	return false;
    }

    
    // Returns an JmxMBeanAttrEntryImpl object for the indicated row.
    // See com.sun.jmx.snmp.agent.SnmpMibTable
    //
    public Object getEntry(SnmpOid oid) 
	throws SnmpStatusException {
	return loadInCache(JmxMBeanContextFactory.getUserData(),null,oid);
    }

    // Loads the indicated MBean  entry in the cache.
    // This has the side effect of:
    // - (re)computing the jmxMBeanTable index
    // - uploading that index into the request-contextual cache
    // - creating the JmxMBeanEntryImpl which is then both stored in the
    //   global cache (JmxMBeanTableMetaImpl.JmxMBeanTableCache) and the
    //   request contextual cache
    //
    private Object loadInCache(Object userData, JmxMBeanTableData tableData,
			       SnmpOid rowOid) throws SnmpStatusException {
	if (rowOid == null)
	    throw new SnmpStatusException(SnmpStatusException.noSuchInstance);
	
	// Get the request contextual cache (userData).
	//
	final Map m = (userData instanceof Map)?((Map)userData):null;

	// We know in the case of this table that the index is an integer,
	// it is thus the first OID arc of the index OID.
	//
	final long   index    = rowOid.getOidArc(0);

	// We're going to use this name to store/retrieve the entry in
	// the request contextual cache.
	//
	// Revisit: Probably better programming to put all these strings
	//          in some interface.
	//
	final String entryTag = ((m==null)?null:("JmxMBeanTable.entry." +
						 index));

	// If the entry is in the cache, simply return it.
	//
	if (m != null) {
	    final Object entry = m.get(entryTag);
	    if (entry != null) return entry;
	}

	// The entry was not in the cache, make a new one.
	//
	// Get the cached data.
	//
	if (tableData == null) tableData = getTableDatas(m);
	
	// tableData should never be null.
	//
	if (tableData == null) 
	    throw new SnmpStatusException(SnmpStatusException.noSuchInstance);

	// Get the data associated with our entry.
	//
	final JmxMBeanTableDataItem data = 
	    (JmxMBeanTableDataItem)tableData.getData(rowOid);

	// data may be null if the OID we were given is not valid.
	// 
	if (data == null) 
	    throw new SnmpStatusException(SnmpStatusException.noSuchInstance);

	// make the new entry (transient object that will be kept only
	// for the duration of the request.
	//
	JmxMBeanEntryImpl entry = data.getEntry();
	if (entry == null) try {
	    entry = 
		new JmxMBeanEntryImpl(myMib,this);
	    entry.JmxMBeanIndex = new Integer((int)index);
	    entry.setJmxMBeanObjectName(data.getObjectName().toString());
	    entry.setJmxMBeanRowStatus(
		  new EnumJmxMBeanRowStatus(EnumRowStatus.active));
	    data.setEntry(entry);
	    if (Boolean.getBoolean("info")) 
		System.err.println("ADDED: JmxMBeanTable["+index+"]="+data);
	} catch (SnmpStatusException x) {
	    if (Boolean.getBoolean("info"))
		System.err.println("Failed to add MBean entry: " + data);
	    if (Boolean.getBoolean("debug")) x.printStackTrace();
	    throw x;
	} catch (Exception x) {
	    if (Boolean.getBoolean("info"))
		System.err.println("Failed to add MBean entry: " + data);
	    if (Boolean.getBoolean("debug")) x.printStackTrace();	    
	    final SnmpStatusException sn = 
		new SnmpStatusException(SnmpStatusException.noSuchInstance);
	    sn.initCause(x);
	    throw sn;
	}

	// Put the entry in the cache in case we need it later while processing
	// the request.
	//
	if (m != null && entry != null) {
	    m.put(entryTag,entry);
	}

	return entry;
    }

    /**
     * This method is invoked when the creation of a new entry is requested
     * by a remote SNMP manager.
     * <br>By default, remote entry creation is disabled - and this method 
     * will not be called. You can dynamically switch the entry creation
     * policy by calling <code>setCreationEnabled(true)</code> and <code>
     * setCreationEnabled(false)</code> on this object.
     * <p><b><i>
     * This method is called internally by the SNMP runtime and you 
     * should never need to call it directly. </b></i>However you might want 
     * to extend it in order to implement your own specific application 
     * behavior, should the default behavior not be at your convenience.
     * </p>
     * <p>
     * @param req   The SNMP  subrequest requesting this creation
     * @param rowOid  The OID indexing the conceptual row (entry) for which 
     *                the creation was requested.
     * @param depth The position of the columnar object arc in the OIDs
     *              from the varbind list.
     *
     * @exception SnmpStatusException if the entry cannot be created.
     */
    public void createNewEntry(SnmpMibSubRequest req, SnmpOid rowOid,
			       int depth) 
	throws SnmpStatusException {
	// create a new MBean
	//
	final long aJmxMBeanIndex = cache.checkIndex(rowOid);

	SnmpVarBind rowStatusVb = req.getRowStatusVarBind();

	if (rowStatusVb == null) 
	    throw new SnmpStatusException(
		      SnmpStatusException.snmpRspNoCreation);

	// Create the new entry
	//
	JmxMBeanEntryImpl entry = new JmxMBeanEntryImpl(myMib,this);
	entry.JmxMBeanIndex = new Integer((int)aJmxMBeanIndex);

	// Perform the necessary initialization step that are needed
	// when a row is created remotely:
	//   * Local creation of row is triggered by the reception of
	//     a MBeanServerNotification: the MBean it mirrors is 
	//     allready registered
	//   * Remote creation of a row is triggered by an SNMP SET request,
	//     and will trigger the creation of a new MBean - whose ObjectName
	//     and Class Name are supplied in the SET request. Creating
	//     and registering this MBean will in turn trigger an 
	//     MBeanServerNotification.
	// This means that special care must be taken so that local/remote
	// creation do not step on each other.
	//
	// The call to createFromRemote() will
	//     1) check the validity of the input parameter,
	//     2) store all the needed information so that the requested
	//        MBean is finally created and registered.
	//     3) ensure that the MBeanServerNotification emitted as a result
	//        of this registration does not create a dupplicated row
	//        at a different index. 
	// 
	entry.createFromRemote(req,rowStatusVb);

	// Get the request contextual cache (userData).
	//
	final Map m = JmxMBeanContextFactory.getUserData();

	// We're going to use this name to store/retrieve the entry in
	// the request contextual cache.
	//
	// Revisit: Probably better programming to put all these strings
	//          in some interface.
	//
	final String entryTag = ((m==null)?null:("JmxMBeanTable.entry." +
						 aJmxMBeanIndex));

	// Put the entry in the cache in case we need it later while processing
	// the request.
	//
	if (m != null && entry != null) {
	    m.put(entryTag,entry);
	}

    }

    /**
     * Remove a table row upon a remote manager request.
     *
     * This method is called internally when <code>getRowAction()</code>
     * yields <i>destroy</i> - i.e.: it is only called when a remote 
     * manager requests the removal of a table row.<br>
     * You should never need to call this function directly.
     * <p>
     * By default, this method simply calls <code>removeEntry(rowOid)
     * </code>.
     * <p>
     * You can redefine this method if you need to implement some 
     * specific behavior when a remote row deletion is invoked.
     * <p>
     * Note that specific checks should not be implemented in this
     * method, but rather in <code>checkRemoveTableRow()</code>. 
     * If <code>checkRemoveTableRow()</code> succeeds and this method 
     * fails afterward, the atomicity of the original SET request can no 
     * longer be guaranteed.
     * <p>
     *
     * @param req    The sub-request that must be handled by this node.
     *
     * @param rowOid The <CODE>SnmpOid</CODE> identifying the table
     *               row involved in the operation.
     *
     * @param depth  The depth reached in the OID tree.
     *
     * @exception SnmpStatusException if the actual row deletion fails.
     *            This should not happen since it would break the 
     *            atomicity of the SET request. Specific checks should
     *            be implemented in <code>checkRemoveTableRow()</code>
     *            if needed. If the entry does not exists, no exception
     *            is generated and the method simply returns.
     *
     **/
    public void removeTableRow(SnmpMibSubRequest req, SnmpOid rowOid,
			       int depth) 
	throws SnmpStatusException {
	final Object entry = getEntry(rowOid);
	final JmxMBeanEntryMBean e = 
	    (JmxMBeanEntryMBean) entry;
	final EnumJmxMBeanRowStatus destroy =
	    new EnumJmxMBeanRowStatus(EnumRowStatus.destroy);
	e.setJmxMBeanRowStatus(destroy);
    }

    /**
     * Return the number of entries in the virtual table.
     **/
    public int getTableSize() {
	return getTableDatas().size();
    }

    //------------------------------------------------------------------------
    // 
    // Remote Management of the jmxMBeanTable table.
    //
    //------------------------------------------------------------------------

    /** 
     * This method is called from JmxMBeanEntryImpl, when a remote SNMP
     * manager trigger the creation of a new MBean through a SET request.
     * @param mbean The just created MBean to be registered.
     * @param name  The ObjectName of the given MBean entry.
     * @param entry The entry in the jmxMBeanAttrTable that mirrors the 
     *              new MBean.
     **/
    synchronized void registerMBean(Object mbean, ObjectName name, 
				    JmxMBeanEntry entry) 
	throws Exception {
	synchronized (cache) {
	    cache.registerName(name,
			       new SnmpOid(entry.JmxMBeanIndex.longValue()));
	    getMibServer().registerMBean(mbean,name);
	}
    }

    /** 
     * This method is called from JmxMBeanEntryImpl, when a remote SNMP
     * manager trigger the removal of a register MBean through a SET request.
     * @param name  The ObjectName of the MBean mirrored by the given entry.
     * @param entry The entry in the jmxMBeanAttrTable that mirrors the MBean
     *              named by <var>name</var>.
     **/
    synchronized void unregisterMBean(ObjectName name, 
				      JmxMBeanEntry entry) 
	throws Exception {
	getMibServer().unregisterMBean(name);
    }

    /**
     * Get the JmxMBeanTableData that holds the jmxMBeanTable data.
     * First look it up in the request contextual cache, and if it is 
     * not found, obtain it from the weak cache.
     * <br>The request contextual cache will be released at the end of the
     * current requests, and is used only to process this request.
     * <br>The weak cache is shared by all requests, and is only
     * recomputed when it is found to be obsolete.
     * Since we will always look in the request contextual cache first
     * we ensure that we will always use the same JmxMBeanTableData for
     * the whole duration of the request. A new incoming request may get
     * the same, or a different JmxMBeanTableData object from the
     * JmxMBeanTableCache.
     * <br>Note that the JmxMBeanTableData put in the request contextual 
     * cache is then used by that request for the whole duration of that 
     * request, no matter how long that request needs to complete, in order 
     * to preserve the coherency of the returned information. This does
     * not prevent concurrent request to get a more up-to-date JmxMBeanTableData
     * object from the JmxMBeanTableCache.
     **/
    public JmxMBeanTableData getTableDatas() {
	final Map userData = JmxMBeanContextFactory.getUserData();
	return getTableDatas(userData);
    }

    /**
     * Get the JmxMBeanTableData that holds the jmxMBeanTable data.
     * First look it up in the request contextual cache, and if it is 
     * not found, obtain it from the weak cache.
     * <br>The request contextual cache will be released at the end of the
     * current requests, and is used only to process this request.
     * <br>The weak cache is shared by all requests, and is only
     * recomputed when it is found to be obsolete.
     * Since we will always look in the request contextual cache first
     * we ensure that we will always use the same JmxMBeanTableData for
     * the whole duration of the request. A new incoming request may get
     * the same, or a different JmxMBeanTableData object from the
     * JmxMBeanTableCache.
     * <br>Note that the JmxMBeanTableData put in the request contextual 
     * cache is then used by that request for the whole duration of that 
     * request, no matter how long that request needs to complete, in order 
     * to preserve the coherency of the returned information. This does
     * not prevent concurrent request to get a more up-to-date JmxMBeanTableData
     * object from the JmxMBeanTableCache.
     * @param userData The request contextual cache.
     **/
    public JmxMBeanTableData getTableDatas(Object userData) {
	final Map m;
	if (userData instanceof Map) m=(Map) userData;
	else m=null;

	// Look in the contextual cache.
	if (m != null) {
	    final JmxMBeanTableData tableData = 
		(JmxMBeanTableData)m.get("JmxMBeanTable.tableData");
	    if (tableData != null) return tableData;
	}

	// No tableData in contextual cache, make a new one.
	final JmxMBeanTableData tableData = cache.getTableDatas(userData);
	
	if (m != null && tableData != null ) 
	    m.put("JmxMBeanTable.tableData",tableData);

	return tableData;
    }

    /**
     * Returns the MBeanServer which is exposed through this MIB.
     **/
    public MBeanServer getMibServer() {
	return  myMib.getMibServer();
    }

}
