/*
 * @(#)file      SimpleManager.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.17
 * @(#)lastedit  04/04/21
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// JDMK import
//
import com.sun.management.snmp.SnmpDefinitions;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.SnmpStatusException;
import com.sun.management.snmp.manager.SnmpPeer;
import com.sun.management.snmp.manager.SnmpParameters;
import com.sun.management.snmp.manager.SnmpRequest;
import com.sun.management.snmp.manager.SnmpSession;
import com.sun.management.snmp.manager.usm.SnmpUsmPeer;
import com.sun.management.snmp.manager.usm.SnmpUsmParameters;
import com.sun.management.snmp.SnmpEngine;
import com.sun.management.snmp.SnmpEngineId;

import com.sun.management.snmp.SnmpEventReportDispatcher;
import com.sun.management.snmp.SnmpOidTableSupport;
import com.sun.jdmk.tasks.DaemonTaskServer;

/**
 * The class is a simple example showing how to send requests to a 
 * Java DMK proxied MIB.
 *
 * <p>When calling the program, you must specify:
 * <br>     - host: hostname of the SNMP proxy you want to query.
 * <br>     - port: port number to use.
 *
 * <p> A new Java DMK java property is used to set a security file:
 * <pre>
 *     -Djdmk.security.file=<i>file</i>
 * </pre>
 * <p>It contains security information related to users. Have a look at the 
 * READMEV3 file in order to find out how SNMP V3 managers are configured.
 */

public class SimpleManager {
   
    /**
     * Main entry point.
     * <p>When calling the program, you must specify:
     * <br> 1) host: hostname of the SNMP agent you want to query.
     * <br> 2) port: port number to use.
     */
    public static void main(String argv[]) {
	SnmpSession session = null;
	
        if (argv.length != 2) {
            usage();
            java.lang.System.exit(1);
        }

	// Check arguments first
	// host and port.
        final String host = argv[0];
        final String port = argv[1];

        try {
            
            // Initialize the SNMP Manager API.
            // Specify the OidTable containing the 4 oids.
            final SnmpOidTableSupport oidTable = new MIB1_MIBOidTable();
            SnmpOid.setSnmpOidTable(oidTable);
	    
            // At any time, it is possible to add knowledge in the OidTable
            // using the SnmpOidDatabaseSupport.
            // For instance, we could start the Manager API like that:
            //    SnmpOidDatabaseSupport oidDB = 
	    //               new SnmpOidDatabaseSupport();
            //    SnmpOid.setSnmpOidTable(oidDB);
            // Then over time, MIB knowledge can be added as follows:
            //    oidDB.add(oidTable);
	  
	    // Build the session. A session creates, controls and manages 
	    // one or more requests.
            //
	    try {
		//When instantiating a session, a new SNMP V3 engine is 
		// instantiated.
		session= new SnmpSession("SimpleManager session");
	    }catch(SnmpStatusException e) {
		println(e.getMessage());
		java.lang.System.exit(0);
	    }
	    catch(IllegalArgumentException e) {
		//If the engine configuration is faulty
		println(e.getMessage());
		java.lang.System.exit(0);
	    }
	
	    // SNMP V3 introduces the notion of SnmpEngine. An engine is 
	    // associated to the session. Other objects  might need the
	    // engine. You can access it using getEngine method.
	    //
	    final SnmpEngine engine = session.getEngine();
	    
	    // Create a SnmpPeer object for representing the entity to
	    // communicate with. 
	    // Java DMK 5.0 introduced a new SnmpPeer that is able to deal 
	    // with user based security parameters. Create it as a Java DMK 4.2 
	    // SnmpPeer and give it the local engine as additional parameter.
	    // A USM peer discovers the engine Id of the SNMP V3 agent it
	    // wishes to communicate with. This engine Id will be used 
	    // when sending requests.
	    //
	    final SnmpUsmPeer agentV3 = 
		new SnmpUsmPeer(engine, host, Integer.parseInt(port));
	    
	    // Create parameters to associate to the entity to
	    // communicate with.
	    // This parameter is used to parametrize SNMP requests.
	    // A user name is provided. This user must appears in the 
	    // jdmk.security file.
	    //
	    final SnmpUsmParameters p = 
		new SnmpUsmParameters(engine, "defaultUser");
	    
	    // Security level is set to authNoPriv. This means that 
	    // authentication is applied to the requests making use of 
	    // these parameters.
	    // This level of security must be in sync with the security 
	    // configuration located in jdmk.security. An authentication 
	    // password/algorithm must be provided. 
	    // If the security level is not in sync, an SnmpStatusException
	    // with a failed status will be thrown when making requests.
	    //
	    p.setSecurityLevel(SnmpDefinitions.authNoPriv);

	    // In SNMP V3 you can specify a contextEngineId. This is 
	    // usually the engine Id the agent discovered. Just reuse it.
	    //
	    p.setContextEngineId(agentV3.getEngineId().getBytes());
	    
	    // The newly created parameter must be associated to the agent.
	    //
	    agentV3.setParams(p);
	
	    
	    // When making authenticated requests, some values must be 
	    // discovered.
	    // This is done calling processUsmTimelinessDiscovery.
	    // If this call is not done to the distant SNMP V3 agent, 
	    // authenticated requests will fail.
	    //
	    agentV3.processUsmTimelinessDiscovery();


	    // We want also to make SNMP V2 requests. Just create an 
	    // SnmpPeer.
	    //
	    final SnmpPeer agent = 
		new SnmpPeer(host, Integer.parseInt(port));
	    
	    
	    // Create a taskServer for processing traps
	    // This is an optional step. However using a DaemonTaskServer
	    // to process incomming PDUs makes it possible to empty
	    // the trap socket faster, thus reducing the hazards
	    // of trap loss.
	    // We set the priority of the DaemonTaskServer to 
	    // Thread.NORM_PRIORITY so that emptying the socket takes 
	    // precedence over trap processing.
	    //
	    final DaemonTaskServer taskServer = new DaemonTaskServer();
	    taskServer.start(Thread.NORM_PRIORITY);
	    
	    // Create a listener and dispatcher for SNMP traps 
	    // (SnmpEventReportDispatcher).
            // SnmpEventReportDispatcher is run as a thread and listens 
	    // for traps in UDP port = agent port + 1.
	    // Add TrapListenerImpl as SnmpTrapListener.
	    // TrapListenerImpl will receive a callback when a valid trap 
	    // PDU is received.
	    // You must provide the engine to the SnmpEventReportDispatcher.
	    //
	    final SnmpEventReportDispatcher trapAgent =
		new SnmpEventReportDispatcher(engine, 
					      Integer.parseInt(port) + 1, 
					      null,taskServer, null);
	    trapAgent.addTrapListener(new TrapListenerImpl());
            final Thread trapThread = new Thread(trapAgent);
	    trapThread.setPriority(Thread.MAX_PRIORITY);
	    trapThread.start();
	    
	    
	    // Build the list of variables you want to query.
	    // For debug purposes, you can associate a name to your list.
	    //
	    final SnmpVarBindList list = 
		new SnmpVarBindList("SimpleManager varbind list");
	    
	    // We want to read the sysDescr1, sysDescr2, sysDescr3, 
	    // sysDescr4 variable.
	    //
            // We will thus query "sysDescr#.0", as sysDescr# are scalar
	    // variables (see RFC 1157, section 3.2.6.3.  Identification 
            // of Object Instances, or RFC 2578, section 7.  Mapping of 
	    // the OBJECT-TYPE macro).
	    //
	    list.addVarBind("sysDescr1.0");
	    list.addVarBind("sysDescr2.0");
	    list.addVarBind("sysDescr3.0");
	    list.addVarBind("sysDescr4.0");
	    

	    // Make the SNMP V3 get request and wait for the result.
	    //
	    println("\n>> Press <Enter> if you want to send " +
		    "a SNMP V3 request.");
	    java.lang.System.in.read();
	    
	    SnmpRequest request = 
		session.snmpGetRequest(agentV3, null, list);
	    println("SimpleManager::main:" +
		    " Send SNMP V3 get request to SNMP V3 agent on " + 
		    host + " at port " + port);
	    boolean completed = request.waitForCompletion(10000);
	    
	    // Check for a timeout of the request.
	    //
            if (completed == false) {
                println("SimpleManager::main:" + 
			" Request timed out. Check reachability of agent");
		
                // Print request.
                //
                println("Request: " + request.toString());
                java.lang.System.exit(0);
            }
	    
            // Now we have a response. Check if the response contains 
	    // an error.
            //
            int errorStatus = request.getErrorStatus();
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                println("Error status = " + 
			SnmpRequest.snmpErrorToString(errorStatus));
                println("Error index = " + request.getErrorIndex());
                java.lang.System.exit(0);
            }
       
            // Now we shall display the content of the result.
            //
            SnmpVarBindList result = request.getResponseVarBindList();
            println("Result: \n" + result);
       

	    //Do another request in SNMP V2.
	    println("\n>> Press <Enter> if you want to send " +
		    "a SNMP V2 request.");
            java.lang.System.in.read();
	    
	    request = session.snmpGetRequest(agent, null, list);
	    println("SimpleManager::main:" + 
		    " Send SNMP V2 get request to SNMP agent on " + 
		    host + " at port " + port);
	    completed = request.waitForCompletion(10000);
	    
	    // Check for a timeout of the request.
	    //
            if (completed == false) {
                println("SimpleManager::main:" +
			" Request timed out. Check reachability of agent");
		
                // Print request.
                //
                println("Request: " + request.toString());
                java.lang.System.exit(0);
            }
	    
            // Now we have a response. Check if the response contains 
	    // an error.
            //
	    errorStatus = request.getErrorStatus();
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                println("Error status = " + 
			SnmpRequest.snmpErrorToString(errorStatus));
                println("Error index = " +
			request.getErrorIndex());
                java.lang.System.exit(0);
            }
	    
            // Now we shall display the content of the result.
            //
	    result = request.getResponseVarBindList();
            println("Result: \n" + result);
	    
            
            // Nicely stop the session
            //
            session.destroySession();
	    
	    // Nicely stop the SnmpEventReportDispatcher.
	    //
	    trapAgent.close();
	    taskServer.terminate();

            //
            // That's all !
            //
            java.lang.System.exit(0);
     
        } catch(Exception e) {
            java.lang.System.err.println("SimpleManager::main:" +
					 " Exception occurred: " + e );
            e.printStackTrace();
        }
    }

    public static void  usage() {
        println("java SimpleManager <master_agent_host> <master_agent_port>");
        println("where");
        println("\t-master_agent_host: hostname or IP address " +
		"of the SNMP agent you want to query.");
        println("\t-master_agent_port: port number to use.");
    }

    /**
     * print/println stuff...
     */

    private final static void println(String msg) {
	java.lang.System.out.println(msg);
    }

    private final static void print(String msg) {
	java.lang.System.out.print(msg);
    }
}
