/*
 * @(#)file      Generator.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.8
 * @(#)lastedit  04/05/10
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

import com.sun.jdmk.contrib.kstat.KernelStat;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Hashtable;

/**
 * The "kstat" example runs only on machines with the Solaris OS.
 * <p>
 * This class automatically generates an MBean for any kernel
 * information that can be accessed using the kstat interface.
 * <p>
 * This class takes as input a module name, a module instance number,
 * and an kstat instance name and generates the associated MBean 
 * interface and MBean implementation.
 * The MBean will make use of the KernelStat class that allows to
 * access the kstat info from Java.
 * <p>
 * This class try to make a best guess in order to name the generated
 * class:
 * <ul>
 *   <li>
 *      if the kstat instance name ends with the module instance number,
 *      (e.g. module="hme", instance=0, name="hme0", or
 *            module="hme", instance=0, name="hmec0")
 *      then the java class name is derived from the kstat instance name
 *      by subtracting the module instance number.<br>
 *      => will generate the class hme.java and interface hmeMBean.java
 *         and the class hmec.java and interface hmecMBean.java
 *      You can then access the statistics of "hme0" by creating the MBean:
 *          <br><code>  hmeMBean hme0 = new hme(0,"hme0");  </code><br>
 *      and access the statistics of "hme1" by creating the MBean:
 *          <br><code>  hmeMBean hme1 = new hme(1,"hme1");  </code><br>
 *   </li>
 *   <li>
 *      Otherwise, the kstat instance name is used as the class name.
 *      (e.g. module="vmem", instance=74, name="semaphore")<br>
 *      => will generate the class semaphore.java and interface 
 *         semaphoreMBean.java.<br>
 *      Since the default constructor will instantiate an MBean using
 *      the same parameters that those that were used when generating the 
 *      class, you can access the statistics for the vmem semaphore by
 *      creating an MBean:
 *           <br><code>  semaphoreMBean sem = new semaphore(); </code><br>
 *      which in this case is strictly equivalent to
 *           <br><code>  semaphoreMBean sem = new semaphore(74,"semaphore"); 
 *           </code><br>
 *   </li>
 *   </ul>
 * <p>
 * In order to start the class, you need to invoke:
 * <p>
 *      java Generator <module-name> <instance-name> <module-instance-number>
 * <p>
 * where:
 * <p>
 *      - <i>module-name</i> is the module name to access (such as "hme")
 *      - <i>instance-name</i> is the specific instance of the module to
 *        address (such as "hme0")
 *      - <i>module-instance-number</i> is the specific module instance 
 *           number (such as 0) in which the instance can be found.
 * <p>
 *      Note that you can omit the instance number, in which case the
 *      first module instance number in which an instance of that name is
 *      found will be used.
 * <p>
 *      You can also omit the instance name, in which case a class/interface
 *      pair will be generated for each instance-name found in the given
 *      module.
 * <p>
 * As an example, the program could be invoked as follows in order to generate
 * an MBean for accessing kernel information on the ethernet driver "hme":
 * <pre>
 *      java Generator hme hme0
 * </pre>
 * In this example, the program will generate a standard MBean called
 * "hme.java" and its associated interface called "hmeMBean.java"; or
 * <pre>
 *      java Generator hme hme0
 * </pre>
 * in which case both hme.java/hmeMBean.java and hmec.java/hmecMBean.java
 * will be generated.
 * <p>
 * In order to run the Generator class access to <code>libstat.so</code>
 * is required in your LD_LIBRARY_PATH. Please make sure to add it to
 * your LD_LIBRARY_PATH after having generated the lib (see README).
 * <p>
 * In order to ease the generation of MBeans a script called "generate_mbean"
 * is provided under examples/kstat.
 * <p>
 * The script takes the same input as the generator program but the script
 * invokes the generator program and then javac for compiling the generated
 * MBeans.
 * <p>
 * Then if you have a Java DMK agent with an HTML adaptor you can try to
 * instantiate the newly created MBeans using a Web browser.
 * <p>
 * To do so click on the Admin button and in the field Java class specify
 * the instance name where instance name could be "hme0".
 * <p>
 * You may also want to instantiate an MBean for "hme1". In that case use
 * the non-default constructor and pass 1 as instance-number and
 * "hme1" as instance-name.
 *
 */
public class Generator {

    /**
     * Create file
     */ 
    private static boolean createFile(String fileName, StringBuffer content) {

        RandomAccessFile file;

        //
        // Check if exist
        //
        File f = new File(fileName);

        if (f.exists()) {
            String sname = fileName + "~";
            File f1 = new File(sname);
            if (f1.exists()) {
                f1.delete();
            }
            f.renameTo(f1);
        }

        try {
            file = new RandomAccessFile(fileName, "rw");
            file.writeBytes(content.toString());
            file.close();
        } catch (IOException e) {
            System.err.println("Cannot create " + fileName +
                               "\n" + e.toString());
            return false;
        }

        System.out.println("Generating: " + fileName);

        return true;
    }

    /**
     * Build a better property name
     */
    private static String bestPropName(String pName) {
        StringBuffer name = new StringBuffer();
        char ch;
        boolean upper = false;

        //
        // Be sure that the name does not contains special characters
        // like % = / + -
        //

        for (int i = 0; i < pName.length(); i++) {
            ch = pName.charAt(i);
            switch (ch) {
            case '=':
            case '%':
            case '/':
            case '+':
            case '-':
            case '*':
            case ' ':
            case ',':
            case '?':
            case ':':
            case '.':
            case '\\':
                //
                // Just remove and put a capital letter on next step
                //
                upper = true;
                break;
            default:
                if (upper) {
                    ch = Character.toUpperCase(ch);
                    upper = false;
                }
                name.append(ch);
                break;
            }
        }

        // name.append(pName);

        return name.toString();
    }

    // Map the data-type string returned by the JNI library
    // to a Java type.
    //
    //   Integer, UnsignedInteger => int
    //   Long, UnsignedLong       => long 
    //   DataChar                 => int[4] (128 bits integers) 
    //   String                   => String
    //   RawData                  => byte[] (not supported)
    //   IntrData                 => int[]  (not supported)
    //   TimerData                => byte[] (not supported)
    //
    private static String buildTypeName(String tName) {
        if (tName == null) 
            throw new IllegalArgumentException("No data type");
        if (tName.equals("Integer")) return "int";
        if (tName.equals("UnsignedInteger")) return "int";
        if (tName.equals("Long")) return "long";
        if (tName.equals("UnsignedLong")) return "long";
        if (tName.equals("DataChar")) return "int[]";
        if (tName.equals("String")) return "String";
        if (tName.equals("RawData")) return "byte[]";
        if (tName.equals("IntrData")) return "int[]";
        if (tName.equals("TimerData")) return "byte[]";
        throw new IllegalArgumentException(tName+ ": unknown data type");
    }


    /**
     * Generate the source files X.java and XMBean.java
     */
    private static boolean buildSources(String mod, int instance, 
                                        String inst, String classname) {

        final String TAB = "    ";
        final String TAB2 = "        ";
        KernelStat ks = new KernelStat(mod);
        StringBuffer buf = new StringBuffer();
        StringBuffer bufIf = new StringBuffer();
        final String clazz = (classname==null?inst:classname);

        // System.out.println("Generating (" + mod + ", " + inst + ")...");

        //
        // Package for this particular example
        //
        buf.append("/**\n * ");
        buf.append("Generated by the build class");
        buf.append("\n */\n\n");
        buf.append("import com.sun.jdmk.contrib.kstat.KernelStat;\n");

        bufIf.append("/**\n * ");
        bufIf.append("Generated by the build class");
        bufIf.append("\n */\n\n");

        buf.append("public class " +
                   clazz +
                   " implements " +
                   clazz +
                   "MBean {\n\n");

        bufIf.append("public interface " +
                     clazz +
                     "MBean {\n\n");

        //
        // Add a KernelStat Element
        //
        buf.append(TAB + "private final KernelStat ks;\n");
        buf.append(TAB + "private final int        modInst;\n");
        buf.append(TAB + "private final String     instName;\n");

        //
        // End of fields
        //
        buf.append("\n");

        //
        // Add a default contructor
        //
        buf.append(TAB  + "public " + clazz + "() {\n" +
                   TAB2 + "this(" + instance
                   + ", \"" + 
                   inst + "\");\n" +
                   TAB  + "}\n\n");

        //
        // Add a non default contructor
        //
        buf.append(TAB  + "public " + clazz + 
                   "(int modInst, String instName) {\n" +
                   TAB2 + "ks = new KernelStat(\"" + mod + "\");\n" +
                   TAB2 + "this.modInst  = modInst;\n" +
                   TAB2 + "this.instName = instName;\n" +
                   TAB  + "}\n\n");

        //
        // Look all the attributes to build the getter
        //
        String prop = ks.statGetProperties(instance, inst);

        if (prop == null) {
            System.err.println("No properties found for (" +
                               mod + ", " + instance + ", " + inst + ")");
            return false;
        }

        // System.out.println("prop = " + prop);

        boolean done = false;
        int index, sep;
        String elt;
        String pName, tName, bName;

        while (!done) {
            index = prop.indexOf(' ');

            if (index >= 0) {
                elt = prop.substring(0, index);
                prop = prop.substring(index + 1).trim();
            } else {
                elt = prop;
                done = true;
            }

            // System.out.println("Processing " + elt);

            sep = elt.indexOf(':');

            if (sep < 0) {
                System.err.println("Invalid element (" + elt + ")");
                continue;
            }

            pName = elt.substring(0, sep);
            tName = elt.substring(sep + 1);

            bName = bestPropName(pName);
            String type = null;
            try {
                type = buildTypeName(tName);
            } catch (IllegalArgumentException e) {
                System.err.println("Warning: Skipping element: " +
                                   "Invalid type name: [" + elt +
                                   "]: " + e.toString());
            }

            if (!"".equals(bName) && type != null) {
                buf.append(TAB + "public " + type + " get" + bName + 
                           "() {\n");
                buf.append(TAB2 + "return ks.statGet" + tName + 
                           "Value(modInst,instName,\"" +
                           pName + "\");\n");
                buf.append(TAB + "}\n\n");

                bufIf.append(TAB + "public " + type + " get" + bName + 
                           "();\n\n");
            }
        }

        //
        // End of file
        //
        buf.append("}");
        bufIf.append("}");

        // System.out.println("\nFile:\n" + buf.toString());
        // System.out.println("\nFile:\n" + bufIf.toString());

        if (!createFile(clazz + ".java", buf))
            return false;
        if (!createFile(clazz + "MBean.java", bufIf))
            return false;

        return true;
    }

    // 
    // Try to guess the best name for the class to generate.
    // module=module-name, eg "hme"
    // modInstance=module-instance-number, eg 0
    // instName=instance-name, eg "hme0"
    // classname=user specified classname - or null.
    //
    public static String bestClassName(String module, int modInstance,
                                       String instName, String classname) {
        // A specific classname is specified in the "classname" property:
        // use it. 
        if (classname != null) return bestPropName(classname);

        // If the instance name ends with the instance number, 
        // take it off...
        //
        final String mi = Integer.toString(modInstance);
        if (instName.endsWith(mi)) {
            final int l = instName.length() - mi.length();
            return bestPropName(instName.substring(0,l));
        }

        // Otherwise, use directly the instance name
        //
        return bestPropName(instName);
    }

    //
    // Arguments is <module name | - > <instance name | - > <module type>
    //
    public static void main(String args[]) {
        if (args.length == 0) {
            System.err.println("Usage:...");
            System.exit(1);
        }

        String modName = null;
        String modInstance = null;
        String instName = null;
        int modType = KernelStat.STAT_NONE;
        String clazz = System.getProperty("classname",null);

        if (!args[0].equals("-")) {
            modName = args[0];
        }

        if (args.length > 1) {
            if (!args[1].equals("-")) {
                instName = args[1];
            }
        }

        if (args.length > 2) {
            if (!args[2].equals("-")) {
                if (args[2].equals("named")) {
                    modType = KernelStat.STAT_NAMED;
                } else if (args[2].equals("io")) {
                    modType = KernelStat.STAT_IO;
                } else if (args[2].equals("intr")) {
                    modType = KernelStat.STAT_INTR;
                } else if (args[2].equals("raw")) {
                    modType = KernelStat.STAT_RAW;
                } else if (args[2].equals("timer")) {
                    modType = KernelStat.STAT_TIMER;
                } else {
                    System.err.println("Invalid type! " +
                                 "(must be: named, io, intr, raw, timer)");
                    System.exit(1);
                }
            }
        }

        //
        // Get the module:instance list concerned
        //
        KernelStat ks;
        String list;

        ks = new KernelStat(modName);
        list = ks.getHeaders(KernelStat.ANY_INSTANCE, instName, modType);

        if (list == null) {
            System.err.println("No elements found");
            System.exit(1);
        }

        System.out.println("Found:\n" + list);

        boolean done = false;
        int index, sep1, sep2;
        String elt;

        final Hashtable classes = new Hashtable(10);
        int classcount = 0;

        while (!done) {
            index = list.indexOf(' ');

            if (index >= 0) {
                elt = list.substring(0, index);
                list = list.substring(index + 1);
            } else {
                elt = list;
                done = true;
            }

            // System.out.println("Processing " + elt);

            sep1 = elt.indexOf(';');
            sep2 = elt.indexOf(';',sep1+1);

            if (sep1 < 0 || sep2 < 0) {
                System.err.println("Invalid element (" + elt + ")");
                continue;
            }

            modName = elt.substring(0, sep1);
            modInstance = elt.substring(sep1+1, sep2);
            instName = elt.substring(sep2 + 1);

            // Get the instance number
            final int instance;
            try {
              instance = Integer.parseInt(modInstance);
            } catch (Exception x) {
                System.err.println(modInstance + 
                                   ": bad module instance number for (" +
                                   modName + "," + instName);
                continue;
            }

            // Classname
            final String classname = 
                bestClassName(modName,instance,instName,clazz);

            // Check that we didn't already generate a class by that name.
            final String kelt = (String)classes.get(classname); 
            if (kelt != null) {
                System.err.println(classname + " already generated for " +
                                   (String)classes.get(classname));
                continue;
            }


            if (!buildSources(modName, instance, instName, classname)) {
                System.err.println("Cannot generate the files for (" +
                                   modName + "," + instName + ")");
                continue;
            }
            
            // remember the classname...
            classes.put(classname,elt);
        }
    }
}
